/*=======================================================
  Author:   R. Muth & B. De Sutter
  Version:  $Id: opt.misc.c,v 1.1.1.1 2000/06/07 12:37:02 brdsutte Exp $
  Comment:  This file contains the code for elementary optimizations
=======================================================*/

#include "global.h"
#include "flags.h"
#include "opcode.h"
#include "functions.h"
#include "regs.h"
#include "system.coff.h"

/*=======================================================
@OPTIMIZATION
Eliminate No-ops
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptEliminateNoops"

extern BOOL OptEliminateNoops(FILE *fp)
{
  INDEX  ibbl;
  INDEX  iins;
  BOOL   change;
  NUMBER kcount=0;
  NUMBER kcounth=0;
  INT64  weight=0;
  BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Noop");
  clock_t time = clock();
    
  FOREACH_LIVE_INS(iins)
    {
      ibbl = AINS_BBL(iins);
        
      if( !AinsHasSideEffect(iins) &&
	  (AINS_REG_C(iins) == REG_ZERO || AINS_REG_C(iins) == REG_FZERO)) 
        {
	  if( verbose )
            {
	      fprintf(fp,"\tKilling: ");
	      PrintIns(fp,iins);
            }
	  kcount++;
	  if( ABBL_IS_HOT(ibbl) )
	    kcounth++;
	  weight += ABBL_COUNT(ibbl);
            
	  AinsKill(iins);
        }
    }


  FOREACH_LIVE_INS(iins)
    {
      ibbl = AINS_BBL(iins);
        
      if( AinsIsNoop(iins) )
        {
	  if( verbose )
            {
	      fprintf(fp,"\tKilling: ");
	      PrintIns(fp,iins);
            }
	  kcount++;
	  if( ABBL_IS_HOT(ibbl) )
	    kcounth++;
	  weight += ABBL_COUNT(ibbl);
            
	  AinsKill(iins);
        }
    }
        
  /* kill useless moves */

  FOREACH_LIVE_INS(iins)
    {
      ibbl = AINS_BBL(iins);
                
      if( AINS_CODE(iins) == I_lda &&
	  AINS_DATA(iins) == 0 &&
	  AINS_REG_B(iins) == AINS_REG_C(iins) )
        {
	  if( verbose )
            {
	      fprintf(fp,"\tKilling: ");
	      PrintIns(fp,iins);
            }
	  kcount++;
	  if( ABBL_IS_HOT(AINS_BBL(iins)) )
	    kcounth++;
	  weight += ABBL_COUNT(ibbl);
            
	  AinsKill(iins);
        }

      if( AINS_CODE(iins) == I_cpys &&
	  AINS_REG_B(iins) == AINS_REG_C(iins) &&
	  AINS_REG_B(iins) == AINS_REG_A(iins) )
        {
	  if( verbose )
            {
	      fprintf(fp,"\tKilling: ");
	      PrintIns(fp,iins);
            }
	  kcount++;
	  if( ABBL_IS_HOT(AINS_BBL(iins)) )
	    kcounth++;
	  weight += ABBL_COUNT(ibbl);
            
	  AinsKill(iins);
        }
    }

  time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
  change = (kcount>0);

  InfoStatus(fp,"\t[OptNoop%c] Killed Noops %d#%d  wt: %ld {%d ms}\n",CHANGE(change),kcount,kcounth,weight,time);

  return change;
}

/*=======================================================
@OPTIMIZATION
Eliminate Unreachable Code .
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptRemoveUnreachableCode"

#define ABBL_MAGIC(i) ABBL_U1(i).doubleword.d1

extern BOOL OptRemoveUnreachableCode(FILE *fp, BOOL report)
{
  BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Reach");
  BOOL   xverbose = HasOption(PRINT_OPT_VERBOSE,"FuncNames");
  BOOL   change = 0;
  INDEX iedg;
  INDEX ibbl;
  INDEX ifun;
  INDEX iins;
  BOOL  flag;
  BOOL  flag2;
  NUMBER rcount;
  NUMBER icount;
  NUMBER bcount;
  NUMBER fcount;
  clock_t time = clock();
    
  FOREACH_LIVE_BBL(ibbl)
    {
      if( AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) )
	{
	  ABBL_MARK(ibbl);
	  ABBL_MAGIC(ibbl) = 1;
	}
      else
	{
	  ABBL_UNMARK(ibbl);
	  ABBL_MAGIC(ibbl) = 0;
	}
    }

  FOREACH_LIVE_FUN(ifun)
    {
      if( AFUN_IS_PSEUDO(ifun))
	AFUN_FLAGS(ifun)|=F_MARKED;
      else
	AFUN_FLAGS(ifun)&=~F_MARKED;
    }
  
  rcount = 0;
  
  do
    {
      flag = 0;
      rcount++;
      FOREACH_LIVE_FUN(ifun)
	{
	  if( !(AFUN_FLAGS(ifun)&F_MARKED) )
	    continue;
	  
	  AFUN_FLAGS(ifun)&=~F_MARKED;
	  
	  do
	    {
	      flag2 = 0;
	      
	      FOREACH_FUN_BBL(ifun,ibbl)
		{
		  if (!ABBL_IS_MARKED(ibbl)) continue;
		  
		  ABBL_UNMARK(ibbl);
		  
		  FOREACH_BBL_SUCC(ibbl,iedg)
		    {
		      INDEX jbbl = AEDG_TAIL(iedg);
		      
		      /* do not follow return edges,
			 the link edge will guarantee the survival */
		      
		      if( AEDG_TYPE(iedg) == ET_FUN_RETURN )
			continue;
		      
		      if( ABBL_MAGIC(jbbl) == 0 )
			{
			  ABBL_MAGIC(jbbl) = 1;
			  ABBL_MARK(jbbl);
			  if (!AEDG_IS_INTER(iedg))
			    {
			      flag2 = 1;
			    }
			  else
			    {
			      AFUN_FLAGS(ABBL_FUN(jbbl))|=F_MARKED;;
			      flag = 1;
			    }
			}
		    }
		}
	    }
	  while (flag2);
	}
    }
  while (flag);

  icount = 0;
  bcount = 0;
  fcount = 0;

    
  FOREACH_LIVE_BBL(ibbl)
    {
      if( ABBL_MAGIC(ibbl) == 0 )
        {
	  /* kill exit bbl  only if first */
	  if( ABBL_TYPE(ibbl) == BT_EXIT &&  !ABBL_IS_FIRST(ibbl) )
	    {
	      FOREACH_BBL_INS(ibbl,iins)
		AinsKill(iins);
	      continue;
	    }

	  bcount++;
	  icount += AbblNInss(ibbl);
	  change = 1;            
#if 0
	  CWARNING( CONFIG_USE_PROFILE && ABBL_COUNT(ibbl) > 0,("unreachable bbl %d has bbl count > 0\n",ibbl));
#endif            
	  if (xverbose && report) {
	    printf("@@@FUNC   %s   ins  %d\n", AFUN_NAME(ABBL_FUN(ibbl)), AbblNInss(ibbl));
	  }
            
	  if (ABBL_TYPE(ibbl)==BT_JUMP)
	    AFUN_AUXFLAGS(ABBL_FUN(ibbl))|=FF_KILLED_JUMPS;
	  AbblKill(ibbl);
	  if( verbose )
            {
	      fprintf(fp,"\tKilling Bbl: %d\n",ibbl);
            }
        }
    }

  FOREACH_LIVE_FUN(ifun)
    {
      ibbl = AFUN_BBL_FIRST(ifun);
      if( VALID(ibbl) ) continue;

      fcount++;
      change = 1;            

      if (xverbose && report) {
	printf("@@@FUNC   %s   ins  %d\n", AFUN_NAME(ifun), AfunNInss(ifun));
      }            

      AfunKill(ifun);
      if( verbose )
        {
	  fprintf(fp,"\tKilling Fun: %d [%s]\n",ifun,AFUN_NAME(ifun));
        }
        

    }

  time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

  InfoStatus(fp,"\t[OptReach%c] Unreachable Funs: %d  Bbls: %d  Inss: %d (Rounds: %d) {%d ms}\n",
	     CHANGE(change),fcount,bcount,icount,rcount,time);
#if 0
  FOREACH_LIVE_FUN(ifun)
    {
      ibbl = AFUN_BBL_FIRST(ifun);
      FOREACH_BBL_PRED(ibbl,iedg)
        {
	  if( !AEDG_IS_INTER(iedg) ) continue;
	  if( AEDG_TYPE(iedg) == ET_HELL ) continue;
	  break;
        }

      if( VALID(iedg) ) continue;

      fprintf(fp,"\tPossibly unreachable fun %d [%s] bbl: %d ins: %d\n",
	      ifun,AFUN_NAME(ifun), AfunNBbls(ifun),AfunNInss(ifun));

    }
#endif

  return change;
  
}

/*=======================================================
@OPTIMIZATION
Eliminate Empty Basic Blocks. 
  Remove empty basic blocks which must be a bt_norm
  basic block by forwarding its incoming edges to the next bbl
  Except if the basic block is a dummy entry basic block
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptUnlinkEmptyBbls"

extern BOOL OptUnlinkEmptyBbls(FILE *fp)
{
  BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Emptybbl");
  BOOL  change;
  BOOL  save;
  INDEX ibbl, iedg;
  NUMBER ecount = 0;
  NUMBER ecounth = 0;
  clock_t time = clock();

  FOREACH_LIVE_BBL(ibbl)
    {
      INDEX next;

      if( AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) ) continue;
      if( ABBL_HAS_INSS(ibbl) ) continue;
      if( ABBL_TYPE(ibbl) == BT_EXIT ) continue;
      if( VALID(AFUN_MASTER(ABBL_FUN(ibbl))) ) continue;

      if( verbose )
        {
	  fprintf(fp,"\t[*]  bbl: %d (%c)\n",ibbl,ABBL_TYPE(ibbl));
        }

        
      CASSERT( ABBL_TYPE(ibbl) == BT_NORM,("empty bbl is not bt_norm\n"));
      next = AEDG_TAIL(ABBL_EDG_NORM(ibbl));
      CASSERT( ABBL_TYPE(next) != BT_EXIT,("next bbl is exit bbl\n"));
      CERROR( AFUN_IS_PSEUDO(ABBL_FUN(next)),("next bbl is in PSEUDO\n"));

      /* there is a bad case of a finite loop inside of the Ots library
	 which we skip hereby */

      if( ABBL_TYPE(next) == BT_PAL )
        {
	  if( AEDG_TAIL(ABBL_EDG_LINK(next)) == ibbl)  continue;
        }
#ifdef SQUEEZE
      /*
       * For code compression, we sometimes insert dummy blocks to
       * avoid certain kinds of inter-procedural edges, e.g., from
       * a BT_SWITCH block to its targets.  These dummy blocks
       * should not be eliminated.
       */
      save = FALSE;
      FOREACH_BBL_PRED(ibbl, iedg) {
	save |= AEDG_IS_INTER(iedg);
      }
      FOREACH_BBL_SUCC(ibbl, iedg) {
	save |= AEDG_IS_INTER(iedg);
      }
      if (save) continue;
#endif /* SQUEEZE */

      if( next == ibbl )
        {
	  WARNING(("skipping infinite loop in %d\n",ibbl));
	  continue;
        }
        
        
      AbblMigrateRelocInfo(ibbl,next);   /* very important */
        
      ecount++;
      if( ABBL_IS_HOT(ibbl) )
	ecounth++;

      /*
	the next statement avoids some warning messages
	if ibbl is the first bbl of a function.
	Otherwise, called would be moved from the first to
	the second bbl of a function.
      */


      if( ABBL_IS_FIRST(ibbl) && ABBL_FUN(ibbl) == ABBL_FUN(next) )
        {
	  AbblUnlink(next);
	  AbblInsertAfter(next,ibbl);
        }
        
      AbblUnlink(ibbl);


      AbblMovePredEdges(ibbl,next);

      /* this is extremely stupid but avoid some asserts
	 because otherwise kill would unlink and already unlinked bbl
      */
        
      AbblAppendFancy(ibbl);
      AbblKill(ibbl);

    }

  time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
  change = (ecount>0);

  InfoStatus(fp,"\t[OptEmptybbl%c] Unlinked Empty bbls: %d#%d {%d ms}\n",CHANGE(change),ecount,ecounth,time);
    
  return change;    
}

/*=======================================================
@OPTIMIZATION
@END
=======================================================*/

extern BOOL OptRemoveHellEdges(FILE *fp)
{
  INDEX ifun;
  INDEX pred;
  INDEX ibbl;
  BOOL verbose  = HasOption(PRINT_OPT_VERBOSE,"Removehelledges");

  BOOL  change;
  NUMBER count = 0;

  FOREACH_LIVE_FUN(ifun)
    {
      if( AFUN_IS_PSEUDO(ifun) ) continue;
      
      ibbl = AFUN_BBL_FIRST(ifun);

      //      if (ifun<=CONFIG_MIN || ifun>=CONFIG_MAX) continue;

      /*      if( verbose )
	{
          fprintf(fp, "%5d %30s %c%c %c%c %c%c \n",ifun,AFUN_NAME(ifun),
                  AFUN_FLAGS(ifun) & F_GPREL32 ? '*' : '_',      ABBL_FLAGS(ibbl) & F_GPREL32 ? '*' : '_',
                  AFUN_FLAGS(ifun) & F_REFQUAD ? '*' : '_',      ABBL_FLAGS(ibbl) & F_REFQUAD ? '*' : '_',
                  AFUN_FLAGS(ifun) & F_REFQUAD_HARD ? '*' : '_', ABBL_FLAGS(ibbl) & F_REFQUAD_HARD ? '*' : '_');
	}
      */

      if( ABBL_FLAGS(ibbl) & F_REFQUAD_HARD ) continue;
      
      FOREACH_BBL_PRED(ibbl,pred)
	{
          if( AEDG_TYPE(pred)==ET_HELL)
	    {
              if( verbose )
		fprintf(fp,"remove hell edge to procedure %d %s\n",ifun,AFUN_NAME(ifun));
              
              AedgKillFancy(pred);
              count++;
	    }
	}

      /*      if( verbose )
	      fprintf(fp,"\n");*/
    }


  change = (count>0);
  
  InfoStatus(fp,"\t[Removehelledges%c]  %d hell edges removed\n",CHANGE(change),count);

  return change;

}


/*=======================================================
=======================================================*/
extern BOOL OptRemoveHellEdges2(FILE *fp)
{
  BOOL change = 0;
  INDEX irel;
  INDEX iins;
  INDEX iedg,ibbl;
  NUMBER count = 0;
  INDEX isec;
  INDEX ifun;

  FOREACH_LIVE_FUN(ifun)
    AFUN_FLAGS(ifun)&=~F_MARKED;

  /* Mark functions who's address is loaded from the lita */

  FOREACH_LIVE_INS(iins)
    {
      if (AINS_FLAGS(iins) & FI_EVAL_C)
	{
	  if (AINS_REG_C(iins)==REG_ZERO ) continue;
	  if (AddressIsCode(AINS_C_EVAL(iins)))
	    {
	      if (AINS_TYPE(iins) == IT_JMP || AINS_TYPE(iins) == IT_BRA) continue;
	      if (!(AFUN_FLAGS(ABBL_FUN(AINS_BBL(Addr2Ins(AINS_C_EVAL(iins))))) & F_KILLED))
		{
		  AFUN_FLAGS(ABBL_FUN(AINS_BBL(Addr2Ins(AINS_C_EVAL(iins))))) |= F_MARKED;
		}
	    }
	}
    }


  FOREACH_SEC(isec)    /* Mark functions who's address is found in other data sections */
    {
      if( ASEC_IS_LITA(isec) ) continue;

      FOREACH_SEC_REL(isec,irel)
	{
	  if( ASEC_AREL_EXTERN(isec,irel) )
	    continue;
	  
	  if(  ASEC_AREL_TYPE(isec,irel) == R_LITERAL )
	    continue;
	  
	  if( ASEC_AREL_SECTION(isec,irel) != R_SN_TEXT &&
	      ASEC_AREL_SECTION(isec,irel) != R_SN_INIT &&
	      ASEC_AREL_SECTION(isec,irel) != R_SN_NULL &&
	      ASEC_AREL_SECTION(isec,irel) != R_SN_FINI )
	    continue;
	  
	  if (ASEC_AREL_TYPE(isec,irel)==R_REFQUAD)
	    if (ReadData64(ASEC_AREL_ADDRESS(isec,irel)))
	      AFUN_FLAGS(ABBL_FUN(AINS_BBL(Addr2Ins(ReadData64(ASEC_AREL_ADDRESS(isec,irel)))))) |= F_MARKED;
	}
    }

    
  FOREACH_LIVE_FUN(ifun) /* For unmarked functions, remove the hell edges */
    {
      
      if (AFUN_FLAGS(ifun) & F_MARKED)
	continue;

      ibbl=AFUN_BBL_FIRST(ifun);
	
      if (ifun>3 && ifun != ABBL_FUN(AINS_BBL(Addr2Ins(SysGetProgramEntry()))))
	{
	  ibbl=AFUN_BBL_FIRST(ifun);
	  FOREACH_BBL_PRED(ibbl,iedg)
	    {
	      if (AEDG_FLAGS(iedg) & F_KILLED) continue;
	      if (AEDG_TYPE(iedg) == ET_FUN_RETURN) continue;
	      if (AFUN_IS_PSEUDO(ABBL_FUN(AEDG_HEAD(iedg))))
		{
		  AedgKillFancy(iedg);
		  count++;
		}
	    }
	}
    }


  change = (count>0);
  
  InfoStatus(fp,"\t[Removehelledges2%c]  %d hell edges removed\n",CHANGE(change),count);

  return change;

}

void OptExitFunction()
{
  INDEX ifun;
  INDEX returnbbl,callbbl;
  INDEX call_edge, link_edge;
  INDEX call_edge2;
  INDEX ibbl,jbbl;
  INDEX iedg,jedg;
  INDEX dummy_bbl;

  FOREACH_LIVE_FUN(ifun)
    if (0==strcmp("exit",AFUN_NAME(ifun)))
      {
	//	printf("function exit found\n");
	break;
      }

  if (!VALID(ifun)) return;

  FOREACH_BBL_SUCC(AFUN_BBL_LAST(ifun),iedg)
    {
      if (AEDG_TYPE(iedg)!=ET_FUN_RETURN) continue;
      returnbbl=AEDG_TAIL(iedg);
      if (ABBL_IS_RELOC(returnbbl)) continue;

      call_edge=iedg-1;

      callbbl=AEDG_HEAD(call_edge);

      link_edge = ABBL_EDG_LINK(callbbl);

      if (VALID(link_edge))
	{
	  //	  printf("hurray, return edge found to normal block %d with count %ld\n",returnbbl,ABBL_COUNT(returnbbl));
	  dummy_bbl=AbblNew();
	  ABBL_FUN(dummy_bbl)=ABBL_FUN(returnbbl);
	  ABBL_TYPE(dummy_bbl)=BT_NORM;
	  AbblAppendFancy(dummy_bbl);
	  AedgKillFancy(link_edge);
	  AedgKillFancy(call_edge);
	  AedgLinkFancy(AedgNewFancy(callbbl,dummy_bbl,ET_FUN_LINK));
	  AedgLinkFancy(AedgNewFancy(callbbl,AFUN_BBL_FIRST(ifun),ET_FUN_CALL));
	  AedgLinkFancy(AedgNewFancy(dummy_bbl,dummy_bbl,ET_NORM));
	}
    }
}
