/* For the purposes of this code, a "region" is defined to be a set
   of basic blocks S demarcated by two blocks B1 and B2, such that:

   (1) B1 dominates every block in S; and
   (2) B2 post-dominates every block in S.

   B1 and B2 are themselves in S.  A region is represented by the pair
   of such blocks B1, B2.
   */

#include "global.h"
#include "flags.h"
#include "functions.h"
#include "opcode.h"
#include "regs.h"

#define MIN(x, y)            (x < y ? x : y)
#define MAX(x, y)            (x < y ? y : x)

static int minregion = 0, maxregion = 0;
static void FactoredRegionAddPreHeader(INDEX ifun);

#define NUM_BUCKETS 1024
static INDEX hashtbl[NUM_BUCKETS];

#define HashTbl(i)  hashtbl[i]

#define FOREACH_FUN_BBL_DOM(ifun,ibbl,jbbl) for (jbbl=ABBL_PRE_DOM(ibbl);VALID(jbbl) && ABBL_FUN(jbbl)==ifun;jbbl=ABBL_PRE_DOM(jbbl))

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintRegion"

extern void PrintRegion(INDEX dom, INDEX postdom)
{
  INDEX ifun, ibbl;
  INDEX cfun;
  ifun = ABBL_FUN(dom);

  CASSERT((ifun==ABBL_FUN(postdom)), 
	  ("dom %d and postdom %d are in different functions!\n", dom, postdom));

  AregionMarkBbls(dom, postdom);

  printf("######################### Region <%d, %d> #########################\n",
	 dom, postdom);
  printf("BBLS:");
  FOREACH_FUN_BBL(ifun, ibbl) {
    if (ABBL_IS_MARKED(ibbl)) {
      printf(" %d", ibbl);
    }
  }
  printf("\n");
  printf("stkAlias = %d, stkArg = %d, stkInvalid = %d\n",
	 (AFUN_HAS_STK_ALIAS(ifun) != 0), (AFUN_HAS_STK_ARG(ifun) != 0), (AFUN_HAS_STK_INVALID(ifun) != 0));
  if (AFUN_HAS_STK_ALIAS(ifun)) {
    printf("stkAliasRegs: ");
    PrintRegSet(AFUN_STK_ALIASES(ifun));
  }
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
  FOREACH_FUN_BBL(ifun, ibbl) {
    if (ABBL_IS_MARKED(ibbl)) {
      PrintBblSuper(stdout, ibbl);
      if (ABBL_TYPE(ibbl) == BT_CALL) {
	cfun = ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)));
	printf("\t[CALLEE]: stkAlias = %d, stkArg = %d, stkInvalid = %d\n",
	       (AFUN_HAS_STK_ALIAS(cfun) != 0), (AFUN_HAS_STK_ARG(cfun) != 0), (AFUN_HAS_STK_INVALID(cfun) != 0));
      }
      PrintSep2(stdout);
    }
  }
  printf("###################################################################\n");
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintSloop"

extern void PrintSloop(INDEX dom, INDEX postdom)
{
  printf("\n%%StartSloop\n");

  PrintRegionSloop(dom, postdom);

  printf("\n%%EndSloop\n");
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintSloopRegion"

extern void PrintRegionSloop(INDEX dom, INDEX postdom)
{
  INDEX ifun, iedg, ibbl;
  ifun = ABBL_FUN(dom);

  CASSERT((ifun==ABBL_FUN(postdom)), 
	  ("dom %d [fun %d] and postdom %d [fun %d] are in different functions!\n", dom, ABBL_FUN(dom), postdom, ABBL_FUN(postdom)));

  ifun = ABBL_FUN(dom);
  printf("%%FUNCTION{%s}{%d}{\n",AFUN_NAME(ifun), ifun);
  printf("  %%CALLSITES{\n");
  FOREACH_BBL_PRED(dom, iedg) {
    printf("    %%CALLER{%s{%d},%d,%d,%ld}\n",
	   AFUN_NAME(ABBL_FUN(AEDG_HEAD(iedg))),
	   ABBL_FUN(AEDG_HEAD(iedg)),
	   AEDG_HEAD(iedg),
	   AEDG_TAIL(iedg),
	   ABBL_COUNT(AEDG_HEAD(iedg)));
  }
  printf("  } # End of callsites\n");
  printf("  %%CHECKED{%s}\n",
	 (AFUN_LOOP_ANALYZABLE(ifun) ? "yes" : "no"));

  FOREACH_FUN_BBL(ifun, ibbl) {
    if (Dominates(dom, ibbl) && PostDominates(postdom, ibbl)) {
      PrintLoopBBL(stdout, ibbl, 2);
    }
  }

  printf("  %%EDGES{");
  FOREACH_FUN_BBL(ifun, ibbl) {
    if (Dominates(dom, ibbl) && PostDominates(postdom, ibbl)) {
      FOREACH_BBL_SUCC(ibbl, iedg) {
	if (!AEDG_IS_INTER(iedg) &&
	    Dominates(dom, AEDG_TAIL(iedg)) && 
	    PostDominates(postdom, AEDG_TAIL(iedg))) {
	  if (AEDG_COUNT(iedg) > 0) {
	    printf("{%d -> %d : %ld} ", 
		   AEDG_FROM(iedg), 
		   AEDG_TAIL(iedg),
		   AEDG_COUNT(iedg));
	  }
	  else {
	    printf("{%d -> %d} ", AEDG_FROM(iedg), AEDG_TAIL(iedg) );
	  }
	}
      }
    }
  }
  printf("  }\n");

  printf("} # End of function %s\n\n", AFUN_NAME(ifun));
}


static void AregionMarkBblsRecPred(INDEX dom, INDEX postdom)
{
  INDEX iedg;
  FOREACH_BBL_PRED(postdom,iedg)
    {
      if (AEDG_IS_INTER(iedg)) continue;
      if (AEDG_HEAD(iedg)==dom) continue;
      if (ABBL_IS_VISITED(AEDG_HEAD(iedg))) continue;
      ABBL_SET_VISITED(AEDG_HEAD(iedg));
      AregionMarkBblsRecPred(dom,AEDG_HEAD(iedg));
    }
}

static void AregionMarkBblsRecSucc(INDEX dom, INDEX postdom)
{
  INDEX iedg;
  FOREACH_BBL_SUCC(dom,iedg)
    {
      if (AEDG_IS_INTER(iedg)) continue;
      if (AEDG_TAIL(iedg)==postdom) continue;
      if (ABBL_IS_VISITED(AEDG_TAIL(iedg))) continue;
      ABBL_SET_VISITED(AEDG_TAIL(iedg));
      AregionMarkBblsRecSucc(AEDG_TAIL(iedg),postdom);
    }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AregionMarkBbls"

extern void AregionMarkBbls(INDEX dom, INDEX postdom)
{
  INDEX ifun;
  INDEX ibbl;
  INDEX iedg;
  BOOL chg;

  ifun = ABBL_FUN(dom);

  FOREACH_FUN_BBL(ifun, ibbl) {
    ABBL_UNMARK(ibbl);
    ABBL_UNSET_VISITED(ibbl);
  }

  if (!AFUN_LOOP_ANALYZABLE(ifun)) {
    return;
  }

  /*
  ABBL_SET_VISITED(dom);
  ABBL_SET_VISITED(postdom);

  AregionMarkBblsRecPred(dom,postdom);

  FOREACH_FUN_BBL(ifun,ibbl)
    if (ABBL_IS_VISITED(ibbl))
      {
	ABBL_MARK(ibbl);
	ABBL_UNSET_VISITED(ibbl);
      }
	
  ABBL_SET_VISITED(dom);
  ABBL_SET_VISITED(postdom);

  AregionMarkBblsRecSucc(dom,postdom);
  
  FOREACH_FUN_BBL(ifun,ibbl)
    {
      if (!ABBL_IS_MARKED(ibbl))
	ABBL_UNSET_VISITED(ibbl);
      if (ABBL_IS_VISITED(ibbl))
	ABBL_MARK(ibbl);
    }
  
  return;
  */

  FOREACH_FUN_BBL(ifun, ibbl) {
    ABBL_UNMARK(ibbl);
    ABBL_UNSET_VISITED(ibbl);
  }

  CASSERT((VALID(dom) && VALID(postdom)), 
	  ("bad arguments: dom = %d, postdom = %d\n", dom, postdom));
  CASSERT(Dominates(dom,postdom),
	  ("%d does not dominate %d [loop-analyzable fun = %d]\n", dom, postdom, AFUN_LOOP_ANALYZABLE(ABBL_FUN(dom))));
  CASSERT(PostDominates(postdom,dom),
	  ("%d does not post-dominate %d [loop-analyzable fun = %d]\n", postdom, dom, AFUN_LOOP_ANALYZABLE(ABBL_FUN(dom))));

  chg = TRUE;
  ABBL_MARK(dom);
  ABBL_MARK(postdom);
  ABBL_SET_VISITED(postdom);

  while (chg) {
    chg = FALSE;
    FOREACH_FUN_BBL(ifun, ibbl) {
      if (ABBL_IS_VISITED(ibbl)) continue;
      if (!ABBL_IS_MARKED(ibbl)) continue;
      if (!PostDominates(postdom, ibbl)) continue;

      ABBL_SET_VISITED(ibbl);
      FOREACH_BBL_SUCC(ibbl, iedg) {
	if (AEDG_IS_INTER(iedg)) continue;
	if (ABBL_IS_MARKED(AEDG_TAIL(iedg))) continue;
	if (!PostDominates(postdom, AEDG_TAIL(iedg))) continue;

	ABBL_MARK(AEDG_TAIL(iedg));
	chg = TRUE;
      }
    }
  }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AregionCallsSetjmp"

extern BOOL AregionCallsSetjmp(INDEX irgn)
{
  INDEX dom, postdom;
  INDEX ifun, cfun;
  INDEX ibbl;

  CASSERT(VALID(irgn), ("bad argument\n"));

  dom = AREGION_DOM(irgn);
  postdom = AREGION_POSTDOM(irgn);

  AregionMarkBbls(dom, postdom);

  ifun = ABBL_FUN(dom);

  FOREACH_FUN_BBL(ifun, ibbl) {
    if (!ABBL_IS_MARKED(ibbl)) continue;
    if (ABBL_TYPE(ibbl) != BT_CALL) continue;

    cfun = ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)));
    if (AfunIsSetjmp(cfun)) return TRUE;
  }

  return FALSE;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AregionSize"

extern void AregionSize(INDEX dom, INDEX postdom, NUMBER *nbbls, NUMBER *nins)
{
  INDEX ifun;
  INDEX ibbl;
  NUMBER num_bbls = 0, num_ins = 0;

  CASSERT((VALID(dom) && VALID(postdom)), 
  	  ("bad arguments: dom = %d, postdom = %d\n", dom, postdom));
  CASSERT(Dominates(dom,postdom),
	  ("%d does not dominate %d\n", dom, postdom));
  CASSERT(PostDominates(postdom,dom),
	  ("%d does not post-dominate %d\n", postdom, dom));

  ifun = ABBL_FUN(dom);

  CASSERT((ifun==ABBL_FUN(postdom)), 
	  ("dom %d and postdom %d are in different functions!\n", dom, postdom));

  AregionMarkBbls(dom, postdom);
  /*
   * now work backwards from postdom and make sure that predecessor blocks
   * have been marked.
   */
  /*  FOREACH_FUN_BBL(ifun, ibbl) {
      ABBL_UNSET_VISITED(ibbl);
      }

      ABBL_SET_VISITED(postdom);
      chg = TRUE;

      while (chg) {
      chg = FALSE;
      FOREACH_FUN_BBL(ifun, ibbl) {
      if (!ABBL_IS_VISITED(ibbl)) continue;
      if (ibbl == dom) continue;

      FOREACH_BBL_PRED(ibbl, iedg) {
      if (AEDG_IS_INTER(iedg)) continue;
      if (ABBL_IS_VISITED(AEDG_HEAD(iedg))) continue;

      if ( !(ABBL_IS_MARKED(AEDG_HEAD(iedg))) ) {
      *nbbls = 0;
      *nins  = 0;
      return;
      }

      ABBL_SET_VISITED(AEDG_HEAD(iedg));
      chg = TRUE;
      }
      }
      }*/
  /*
   * finally, count up the no. of bbls and instructions
   */
  FOREACH_FUN_BBL(ifun, ibbl) {
    if (ABBL_IS_MARKED(ibbl)) {
      num_bbls++;
      num_ins += AbblNInss(ibbl);
    }
  }

  *nbbls = num_bbls;
  *nins  = num_ins;
}

/*=======================================================
  The "fingerprint" for a region is a 64-bit value, defined
  as follows:

  - the top 16 bits give the size of the region: the top
    8 bits give the no. of basic blocks, the next 8 bits the
    no. of instructions in the region.  A value that is too
    large to fit in 8 bits is represented by 0xff.

  - The remaining 48 bits encode the type and size of the
    first 8 basic blocks (in a pre-order traversal starting
    at the top, TRUE edges of a conditional being traversed 
    before FALSE edges).  Each block is encoded using 6 bits:
    2 bits for the type, and 4 bits for the no. of instructions.

=======================================================*/
#undef __PROC__
#define __PROC__ "AregionCode"

extern UINT64 AregionCode(INDEX dom, INDEX postdom)
{
  INDEX ifun, ibbl;
  INDEX bbls[8];
  UINT8 bblcode;
  NUMBER nbbls, nins;
  UINT64 code;
  int i, j;

  bbls[0] = dom;
  for (i=1; i<8; i++) bbls[i] = INVALID;

  ifun = ABBL_FUN(dom);
  i = 1; 
  for (j = 0; j < 8; j++) {
    ibbl = bbls[j];
    if (!VALID(ibbl)) break;

    switch (ABBL_TYPE(ibbl)) {
    case BT_NORM:
      if (i < 8) bbls[i++] = AEDG_TAIL(ABBL_EDG_NORM(ibbl));
      break;
    case BT_CALL:
    case BT_PAL:
      if (ibbl < 8) bbls[i++] = AEDG_TAIL(ABBL_EDG_LINK(ibbl));
      break;
    case BT_BRANCH:
      if (ibbl < 8) bbls[i++] = AEDG_TAIL(ABBL_EDG_TRUE(ibbl));
      if (ibbl < 8) bbls[i++] = AEDG_TAIL(ABBL_EDG_FALSE(ibbl));
      break;
    default:
      continue;
    }
  }

  code = 0;
  for (i = 0; i < 8; i++) {
    ibbl = bbls[i];
    if (!VALID(ibbl)) break;

    switch (ABBL_TYPE(ibbl)) {
    case BT_NORM:
      bblcode = 0x1;
      break;
    case BT_CALL:
      bblcode = 0x2;
      break;
    case BT_BRANCH:
      bblcode = 0x3;
      break;
    default:
      bblcode = 0x0;
      break;
    }

    nins = ABBL_NINS(ibbl);
    if ( nins > 15 ) nins = 0xf;

    bblcode = (bblcode << 4) | nins;
    code = (code << 6) | bblcode;
  }
    
  AregionSize(dom, postdom, &nbbls, &nins);

  nbbls = MIN(nbbls, 0xff);
  nins  = MIN(nins,  0xff);

  code |= ( ((UINT64)nbbls << 56) | ((UINT64)nins << 48) );

  return code;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ComputeRegions"

extern void ComputeRegions()
{
  INDEX ifun, ibbl, jbbl;
  INDEX iregion;
  INDEX dom = INVALID, postdom = INVALID;
  INDEX tmp_post_dom;
  BOOL found_new;
  BOOL after_break;
  BOOL found;
  int i, hashval;
  NUMBER n = 0;
  NUMBER nbbls, nins;
  clock_t time = clock();

  ComDomsAndLoops(stdout);
  ComBblNInss();

  for (i = 0; i < NUM_BUCKETS; i++) {
    HashTbl(i) = INVALID;
  }

  FOREACH_LIVE_FUN(ifun) {
    if ( !AFUN_LOOP_ANALYZABLE(ifun) ) continue;
    
    //    printf("going for fun %d with %d bbls \n",ifun,AfunNBbls(ifun));

    FOREACH_FUN_BBL(ifun, ibbl) {
      if ( !(ABBL_NINS(ibbl) > 0) ) continue;
      if (ABBL_TYPE(ibbl) == BT_BRANCH) continue;

      //      printf("going for bbl %d\n",ibbl);

      tmp_post_dom = ibbl;
      after_break = FALSE;

      if (ABBL_TYPE(ibbl) == BT_CALL) continue;
      if (ABBL_IS_LOOP_HEADER(ibbl)) continue;

      FOREACH_FUN_BBL_DOM(ifun,ibbl, jbbl) {
	
	found_new = FALSE;
	
	if (ABBL_POST_DOM(jbbl)==tmp_post_dom)
	  {
	    tmp_post_dom=jbbl;
	    found_new = TRUE;
	  }
	else
	  {
	    after_break = TRUE;
	  }
	
	if ( !(ABBL_NINS(jbbl) > 0) ) continue;
	if (ABBL_TYPE(jbbl) == BT_BRANCH) continue;
	

	
	found = FALSE;
	if (found_new)
	  {
	    if (ABBL_TYPE(ibbl) == BT_CALL
		&& AINS_CODE(ABBL_INS_LAST(ibbl)) == I_jsr) continue;
	    
	    dom = jbbl;
	    postdom = ibbl;
	    found = TRUE;
	  }

#if 1
	if (ABBL_TYPE(postdom) == BT_CALL) continue;
#endif
	if (found) {
	  AregionSize(dom, postdom, &nbbls, &nins);
	  if (nbbls < 2) continue;  /* single-block regions already factored */

	  iregion = AregionNew();
	  if (!VALID(iregion)) {
	    goto Done;  /* no more space for any new regions */
	  }

	  AREGION_NEXT(iregion) = AFUN_REGION(ifun);
	  AFUN_REGION(ifun) = iregion;

	  AREGION_DOM(iregion) = dom;
	  AREGION_POSTDOM(iregion) = postdom;
	  AREGION_NBBLS(iregion) = nbbls;
	  if (nbbls < minregion) minregion = nbbls;
	  if (nbbls > maxregion) maxregion = nbbls;
	  AREGION_NINS(iregion) = nins;
	  AREGION_CODE(iregion) = AregionCode(dom, postdom);
	  hashval = ( (ABBL_CODE(dom) ^ ABBL_CODE(postdom)) + nbbls + nins) & 0x3ff;
	  AREGION_HASH_LINK(iregion) = HashTbl(hashval);
	  HashTbl(hashval) = iregion;
	  n++;
	}
      }
    }
  }

#if 0
  {
    NUMBER n, hash_bucket;
    NUMBER zero, max, maxi, tot;
    INDEX i;

    max = zero = maxi = tot = 0;

    printf("@@@ HASH TABLE DISTRIBUTION:\n@HASH:   ");
    for (hash_bucket = 0; hash_bucket < NUM_BUCKETS; hash_bucket += 1) {
      n = 0;
      for (i = HashTbl(hash_bucket); VALID(i); i = AREGION_HASH_LINK(i)) {
	tot++;
	n++;
      }
      printf("%d; ", n);
      if (n == 0) zero++;
      if (n > max) { max = n; maxi = hash_bucket; }
      if (hash_bucket > 0 && hash_bucket % 20 == 0) printf("\n@HASH:   ");
    }
    printf("\n@@@ HASH: Zero = %d [out of %d], Max = %d, ave = %d\n", 
	   zero, NUM_BUCKETS, max, tot/NUM_BUCKETS);
    printf("--- ");
#if 0
    for (i = HashTbl(maxi); VALID(i); i = ABBL_HASH_LINK(i)) {
      printf("%lx  ", ABBL_CODE(i));
      if (i > 0 && i % 5 == 0) printf("\n--- ");
    }
#endif

    printf("@HASH: ------------------------------------------------------------\n");
  }
#endif

 Done:
  time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
  InfoStatus(stdout,"\t[ComputeRegions]: %d regions [max %d] { %d ms }\n", n, CONFIG_NREGION, time);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ASubregionIdentical"

extern BOOL ASubregionIdentical(INDEX ibbl, INDEX jbbl, INDEX istop, INDEX jstop)
{
  INDEX iedg, jedg;
  INDEX xbbl, ybbl;

  if (ABBL_IS_MARKED(ibbl) && ABBL_IS_MARKED(jbbl)) return TRUE;
  if (ABBL_IS_MARKED(ibbl) || ABBL_IS_MARKED(jbbl)) return FALSE;
  if (ABBL_CODE(ibbl) != ABBL_CODE(jbbl)) return FALSE;
  if (ABBL_NINS(ibbl) != ABBL_NINS(jbbl)) return FALSE;
  if ( !AbblIdentical(ibbl, jbbl) ) return FALSE;
  if (ibbl == istop && jbbl == jstop) return TRUE;
  if (ABBL_TYPE(ibbl) != ABBL_TYPE(jbbl) ||
      AbblNsuccs(ibbl) != AbblNsuccs(jbbl)) return FALSE;
  /*
   * At this point we know that ibbl and jbbl are identical.  We
   * consider their successors, pairwise.
   */
  ABBL_MARK(ibbl);
  ABBL_MARK(jbbl);

  switch (ABBL_TYPE(ibbl)) {
  case BT_RET:
  case BT_EXIT:
    return FALSE;

  case BT_NORM:
    xbbl = AEDG_TAIL(ABBL_EDG_NORM(ibbl));
    ybbl = AEDG_TAIL(ABBL_EDG_NORM(jbbl));
    return ASubregionIdentical(xbbl, ybbl, istop, jstop);

  case BT_CALL:
  case BT_PAL:
    xbbl = AEDG_TAIL(ABBL_EDG_LINK(ibbl));
    ybbl = AEDG_TAIL(ABBL_EDG_LINK(jbbl));
    return ASubregionIdentical(xbbl, ybbl, istop, jstop);

  case BT_BRANCH:
    xbbl = AEDG_TAIL(ABBL_EDG_TRUE(ibbl));
    ybbl = AEDG_TAIL(ABBL_EDG_TRUE(jbbl));
    if (!ASubregionIdentical(xbbl, ybbl, istop, jstop)) {
      return FALSE;
    }
    xbbl = AEDG_TAIL(ABBL_EDG_FALSE(ibbl));
    ybbl = AEDG_TAIL(ABBL_EDG_FALSE(jbbl));
    return ASubregionIdentical(xbbl, ybbl, istop, jstop);
      
  case BT_SWITCH:
    for (iedg = ABBL_SUCC(ibbl), jedg = ABBL_SUCC(jbbl);
	 VALID(iedg) && VALID(jedg); 
	 iedg = AEDG_SUCC(iedg), jedg = AEDG_SUCC(jedg)) {
      xbbl = AEDG_TAIL(iedg);
      ybbl = AEDG_TAIL(jedg);
      if (!ASubregionIdentical(xbbl, ybbl, istop, jstop)) {
	return FALSE;
      }
    }
    if (!VALID(iedg) && !VALID(jedg)) {
      return TRUE;
    }
    else {
      return FALSE;
    }

  default:
    ERROR(("unexpected basic block %d [type %c]\n", ibbl, ABBL_TYPE(ibbl)));
    return FALSE;
  }
  /* control should not reach here */
  return FALSE; 
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AregionIdentical"

extern BOOL AregionIdentical(INDEX irgn, INDEX jrgn)
{
  INDEX idom, ipostdom, jdom, jpostdom;
  INDEX ibbl, jbbl;
  INDEX ifun, jfun;

  if (AREGION_NBBLS(irgn) != AREGION_NBBLS(jrgn) ||
      AREGION_NINS(irgn) != AREGION_NINS(jrgn)) {
    return FALSE;
  }

  idom = AREGION_DOM(irgn);
  ipostdom = AREGION_POSTDOM(irgn);
  jdom = AREGION_DOM(jrgn);
  jpostdom = AREGION_POSTDOM(jrgn);

  if ( (Dominates(idom, jdom) && !Dominates(ipostdom, jdom)) ||
       (Dominates(jdom, idom) && !Dominates(jpostdom, idom)) ) {
    return FALSE;
  }
  
  /* these tests are inserted to speed-up the code: the unmarking of blocks 
     just takes to long */

  if (ABBL_CODE(idom) != ABBL_CODE(jdom)) return FALSE;
  if (ABBL_NINS(idom) != ABBL_NINS(jdom)) return FALSE;
  if ( !AbblIdentical(idom, jdom) ) return FALSE;
  if (idom == ipostdom && jdom == jpostdom) return TRUE;
  if (ABBL_TYPE(idom) != ABBL_TYPE(jdom) ||
      AbblNsuccs(idom) != AbblNsuccs(jdom)) return FALSE;

  ifun = ABBL_FUN(idom);
  jfun = ABBL_FUN(jdom);

  FOREACH_FUN_BBL(ifun, ibbl) {
    ABBL_UNMARK(ibbl);
  }
  if (ifun != jfun) {
    FOREACH_FUN_BBL(jfun, jbbl) {
      ABBL_UNMARK(jbbl);
    }
  }

  return ASubregionIdentical(idom, jdom, ipostdom, jpostdom);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "GroupIdenticalRegions"

extern void GroupIdenticalRegions()
{
  int hash_bucket;
  INDEX irgn, jrgn, xrgn;
  INDEX mi, mj, master, slave;
  NUMBER nrgn = 0;
  clock_t time = clock();

  for (hash_bucket = 0; hash_bucket < NUM_BUCKETS; hash_bucket += 1) {
    for (irgn = HashTbl(hash_bucket); VALID(irgn); irgn = AREGION_HASH_LINK(irgn)) {
      CASSERT(AFUN_LOOP_ANALYZABLE(ABBL_FUN(AREGION_DOM(irgn))),
	      ("region <%d,%d> in un-loop-analyzable function %d\n", AREGION_DOM(irgn), AREGION_POSTDOM(irgn), ABBL_FUN(AREGION_DOM(irgn))));
      if (VALID(AREGION_MASTER(irgn))) continue;
      
      for (jrgn = AREGION_HASH_LINK(irgn); VALID(jrgn); jrgn = AREGION_HASH_LINK(jrgn)) {
	if (AREGION_CODE(irgn) != AREGION_CODE(jrgn)) continue;
	if (AREGION_NBBLS(irgn) != AREGION_NBBLS(jrgn)) continue;
	if (AREGION_NINS(irgn) != AREGION_NINS(jrgn)) continue;
	if (VALID(AREGION_MASTER(jrgn))) continue;
	
	if (AregionIdentical(irgn, jrgn)) {
	  mi = VALID(AREGION_MASTER(irgn)) ? AREGION_MASTER(irgn) : irgn;
	  mj = VALID(AREGION_MASTER(jrgn)) ? AREGION_MASTER(jrgn) : jrgn;
	  master = MAX(mi, mj);
	  slave = MIN(mi, mj);
	  AREGION_MASTER(master) = AREGION_MASTER(slave) = master;
      
	  for (xrgn = master; VALID(AREGION_LINK(xrgn)); xrgn = AREGION_LINK(xrgn)) {
	  }   /* skip to end of list */
	  AREGION_LINK(xrgn) = slave;
	  nrgn++;
#if 0
	  printf("### REGION [%c]: <%d,%d> & <%d,%d> [m: <%d,%d>] [ %d bbls, %d ins ]; [als,arg,inv] = [%d%d%d]\n", 
		 ABBL_TYPE(AREGION_POSTDOM(irgn)),
		 AREGION_DOM(irgn), AREGION_POSTDOM(irgn), 
		 AREGION_DOM(jrgn), AREGION_POSTDOM(jrgn), 
		 AREGION_DOM(master), AREGION_POSTDOM(master), 
		 AREGION_NBBLS(irgn), AREGION_NINS(irgn),
		 (AFUN_HAS_STK_ALIAS(ABBL_FUN(AREGION_DOM(jrgn))) != 0),
		 (AFUN_HAS_STK_ARG(ABBL_FUN(AREGION_DOM(jrgn))) != 0),
		 (AFUN_HAS_STK_INVALID(ABBL_FUN(AREGION_DOM(jrgn))) != 0));
#endif
	}
      }
    }
  }

  time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
  InfoStatus(stdout,"\t[GroupIdenticalRegions]: %d pairs { %d ms }\n", nrgn, time);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AregionRepairPostDomCalls"

extern void AregionRepairPostDomCalls(INDEX irgn)
{
  INDEX jrgn;
  INDEX dom, postdom;
  INDEX ibbl, jbbl;
  INDEX iedg;
  NUMBER n;

  CASSERT(VALID(irgn), ("bad argument\n"));
  CASSERT(VALID(AREGION_MASTER(irgn)),
	  ("invalid master: region <%d,%d>\n", AREGION_DOM(irgn), AREGION_POSTDOM(irgn)));
  CASSERT(ABBL_TYPE(AREGION_POSTDOM(irgn)) == BT_CALL,
	  ("post-dom not of type BT_CALL: region <%d,%d>\n",AREGION_DOM(irgn), AREGION_POSTDOM(irgn)));

  for (jrgn = AREGION_MASTER(irgn); VALID(jrgn); jrgn = AREGION_LINK(jrgn)) {
    dom = AREGION_DOM(jrgn);
    postdom = AREGION_POSTDOM(jrgn);

    n = 0;
    ibbl = INVALID;
    FOREACH_BBL_PRED(AEDG_TAIL(ABBL_EDG_LINK(postdom)), iedg) {
      if (AEDG_IS_INTER(iedg)) continue;
      ibbl = AEDG_TAIL(iedg);
      n++;
    }
    if (n > 1) {
      AREGION_MASTER(jrgn) = INVALID;  /* can we improve on this? */
      continue;
    }

    CASSERT(VALID(ibbl), ("couldn't find return bbl?!\n"));

    jbbl = InsertBblBeforeFancy(ibbl);
    AREGION_POSTDOM(jrgn) = jbbl;
#if 0
    printf("@@@ REPAIR: region <%d,%d> --> %d\n", dom, postdom, jbbl);
#endif
  }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunCallFreq"

extern NUMBER AfunCallFreq(INDEX ifun)
{
  INDEX iedg;
  NUMBER n = 0;

  CASSERT(VALID(ifun), ("bad argument\n"));

  if (AFUN_IS_PSEUDO(ifun)) return INVALID;

  FOREACH_BBL_PRED(AFUN_BBL_FIRST(ifun), iedg) {
    if (!AEDG_IS_INTER(iedg)) continue;
    if (AEDG_TYPE(iedg) == ET_HELL) continue;

    n += ABBL_COUNT(AEDG_HEAD(iedg));
  }

  return n;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AregionFlowFrac"

double AregionFlowFrac(INDEX irgn)
{
  NUMBER tot = 0;
  INDEX jrgn;

  CASSERT(VALID(irgn), ("bad argument\n"));

  for (jrgn = AREGION_MASTER(irgn); VALID(jrgn); jrgn = AREGION_LINK(jrgn)) {
    tot += ABBL_COUNT(AREGION_DOM(jrgn));
  }

  if (ABBL_COUNT(AREGION_DOM(irgn)) > 0) {
    return (double)tot/(double)ABBL_COUNT(AREGION_DOM(irgn));
  }
  else return 0.0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "CopyRegion2Fun"

#define ABBL_MAGIC(i) ABBL_U1(i).doubleword.d1

extern INDEX CopyRegion2Fun(INDEX irgn)
{
  INDEX ifun, xfun;
  INDEX ibbl, xbbl;
  INDEX dom, postdom;
  INDEX iedg, jedg, targetbbl;
  INDEX itab, jtab;
  double flowfrac;

  CASSERT(VALID(irgn),("bad argument\n"));

  dom = AREGION_DOM(irgn);
  postdom = AREGION_POSTDOM(irgn);
  CASSERT((ABBL_FUN(dom)==ABBL_FUN(postdom)), ("weird\n"));
  CASSERT(AFUN_LOOP_ANALYZABLE(ABBL_FUN(dom)), 
	  ("region <%d,%d> in un-loop-analyzable function %d\n", dom, postdom, ABBL_FUN(dom)));

  flowfrac = AregionFlowFrac(irgn);

  ifun = ABBL_FUN(dom);
  /*
   * First, mark the blocks that belong to the region
   */

  AregionMarkBbls(dom, postdom);

  FOREACH_FUN_BBL(ifun, ibbl) {
    ABBL_MAGIC(ibbl) = INVALID;
  }

  xfun = AfunNewFancy();
  AFUN_NAME(xfun) = NewFunName("_sFactorRegion", dom);
  AFUN_FLAGS(xfun) = AFUN_FLAGS(ifun); /* NEEDS WORK */

  FOREACH_FUN_BBL(ifun, ibbl) {
    if ( !(ABBL_IS_MARKED(ibbl)) ) continue;

    if (ABBL_IS_LAST(ibbl)) {
      ABBL_MAGIC(ibbl) = AFUN_BBL_LAST(xfun);
      continue;
    }
    xbbl = AbblDup(ibbl);
    ABBL_FUN(xbbl) = xfun;
    ABBL_MAGIC(ibbl) = xbbl;
    if (ibbl == dom) {
      AbblAppendFirst(xbbl);
    }
    else {
      AbblAppendFancy(xbbl);
    }
    
    if( ABBL_FLAGS(ibbl) & F_GPREL32 ) {
      ABBL_FLAGS(xbbl) |= F_GPREL32;
      AINS_FLAGS(ABBL_INS_FIRST(xbbl)) |= F_GPREL32;
    }
    
    ABBL_COUNT(xbbl) = flowfrac * ABBL_COUNT(ibbl);
  }
  /*
   * duplicate edges
   */
  FOREACH_FUN_BBL(ifun,ibbl) {
    if ( !(ABBL_IS_MARKED(ibbl)) ) continue;
    if (!VALID(ABBL_MAGIC(ibbl))) continue;
    if (ibbl == postdom) continue;
    
    /* invariant: link edge before call edge */
    if( ABBL_TYPE(ibbl) == BT_CALL || ABBL_TYPE(ibbl) == BT_PAL ) {
      if( AbblNsuccs(ibbl) != 2 ) {
	ERROR(("number of succs for call != 2\n"));
	PrintBblSuper(stdout,ibbl);
	ERROR(("number of succs for call != 2\n"));
      }
      
      iedg = ABBL_EDG_LINK(ibbl);
      if (!ABBL_IS_MARKED(AEDG_TAIL(iedg))) continue;

      CASSERT(VALID(ABBL_MAGIC(AEDG_TAIL(iedg))), ("link edge not followed? [ibbl %d]\n", ibbl));
      jedg = AedgNewFancy(ABBL_MAGIC(AEDG_HEAD(iedg)),
			  ABBL_MAGIC(AEDG_TAIL(iedg)),AEDG_TYPE(iedg));
      AedgLinkFancy(jedg);
      AEDG_FLAGS(jedg) = AEDG_FLAGS(iedg);
      
      iedg = ABBL_EDG_CALL(ibbl);
      jedg = AedgNewFancy(ABBL_MAGIC(AEDG_HEAD(iedg)),AEDG_TAIL(iedg),AEDG_TYPE(iedg));
      AedgLinkFancy(jedg);
      AEDG_FLAGS(jedg) = AEDG_FLAGS(iedg);
    }
    else {
      FOREACH_BBL_SUCC(ibbl,iedg) {
	if( AEDG_TYPE(iedg) == ET_FUN_RETURN ) continue;
	if( AEDG_TYPE(iedg) == ET_COMPENSATE ) continue;
	if( AEDG_IS_INTER(iedg) ) {
	  ERROR(("unexpected inter edge: %d --> %d\n", AEDG_HEAD(iedg), AEDG_TAIL(iedg)));
	  PrintEdgSuper(stdout,iedg);
	}

	targetbbl = ABBL_MAGIC(AEDG_TAIL(iedg));
	CASSERT(VALID(targetbbl), ("region closure property violated: %d -> %d [fun %d, region %d <%d,%d>]\n",
				   AEDG_HEAD(iedg), AEDG_TAIL(iedg), ifun, irgn, dom, postdom));
	jedg = AedgNewFancy(ABBL_MAGIC(AEDG_HEAD(iedg)), targetbbl,AEDG_TYPE(iedg));
	AedgLinkFancy(jedg);
	AEDG_FLAGS(jedg) = AEDG_FLAGS(iedg);    
      }
    }
  }
  /*
   * duplicate jump tables and patch the bbl
   */
  FOREACH_FUN_BBL(ifun,ibbl) {
    if ( !(ABBL_IS_MARKED(ibbl)) ) continue;
    if( ABBL_TYPE(ibbl) != BT_SWITCH ) continue;

    xbbl = ABBL_MAGIC(ibbl);
    CASSERT( VALID(ABBL_JUMP_TAB(ibbl)),("jump tab invalid\n") );
    CASSERT( !VALID(ABBL_JUMP_TAB(xbbl)),("jump tab valid\n") );
    
    FOREACH_BBL_TAB(ibbl,itab) {
      jtab = AtabNew();
      targetbbl = ATAB_TARGET(itab);
      if (VALID(ABBL_MAGIC(targetbbl))) {
	targetbbl = ABBL_MAGIC(targetbbl);
      }
      ATAB_TARGET(jtab) = targetbbl;
      ATAB_DATA_ADDRESS(jtab)=ATAB_DATA_ADDRESS(itab);
      if( !VALID(ABBL_JUMP_TAB(xbbl)) ) {
	ABBL_JUMP_TAB(xbbl) = jtab;    
      }        
    }
    
    /* add terminator */
    jtab = AtabNew();
    ATAB_TARGET(jtab) = INVALID;
    ATAB_DATA_ADDRESS(jtab) = INVALID;
  }
  
  ibbl = ABBL_MAGIC(postdom);
  xbbl = AFUN_BBL_LAST(xfun);

  ABBL_TYPE(ibbl) = BT_RET;

  AedgLinkFancy( AedgNewFancy(ibbl, xbbl, ET_EXIT_LINK));

  return xfun;
}

/*=======================================================
Shift each stack reference in the function ifun by delta bytes.
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunShiftStackRefs"

extern void AfunShiftStackRefs(INDEX ifun, NUMBER delta)
{
  INDEX ibbl;
  INDEX iins;

  FOREACH_FUN_BBL(ifun, ibbl) {
    FOREACH_BBL_INS(ibbl, iins) {
      switch (AINS_CODE(iins)) {
      case I_ldq_u:
      case I_ldwu:
      case I_ldf:
      case I_ldg:
      case I_lds:
      case I_ldt:
      case I_ldq:
      case I_ldl:
	if (AINS_REG_B(iins) != REG_SP) continue;
	AINS_DATA(iins) = AINS_DATA(iins) + delta;
	break;

      case I_stw:
      case I_stb:
      case I_stq_u:
      case I_stf:
      case I_stg:
      case I_sts:
      case I_stt:
      case I_stl:
      case I_stq:
	if (AINS_REG_A(iins) != REG_SP) continue;
	AINS_DATA(iins) = AINS_DATA(iins) + delta;
	break;
      }
    }
  }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunHasForbiddenLoads"

extern BOOL AfunHasForbiddenLoads(INDEX ifun, REGSET regs)
{
  INDEX ibbl;
  INDEX iins;

  CASSERT(VALID(ifun), ("invalid function argument\n"));

  if (AFUN_IS_PSEUDO(ifun)) return TRUE;
  if (REGSET_INTERSECT(regs, REGS_INT) == REGS_NONE) return FALSE;

  FOREACH_FUN_BBL(ifun, ibbl) {
    FOREACH_BBL_INS(ibbl, iins) {
      if (!(AINS_IS_LOAD(iins) || AINS_IS_STORE(iins)) ) continue;

      if (REGSET_IS_MEMBER(AINS_REG_B(iins), regs)) return TRUE;
    }
  }

  return FALSE;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ARegionIsWellBehaved"

extern BOOL ARegionIsWellBehaved(INDEX irgn)
{
  INDEX dom, postdom;
  INDEX ifun, cfun;
  INDEX ibbl;
  INDEX iins;

  if (AregionCallsSetjmp(irgn)) return FALSE;

  dom = AREGION_DOM(irgn);
  postdom = AREGION_POSTDOM(irgn);
  ifun = ABBL_FUN(dom);

  if (AFUN_HAS_STK_INVALID(ifun) != 0) return FALSE;
#if 0
  if (AFUN_HAS_STK_ARG(ifun) != 0) return FALSE;
#endif

  AregionMarkBbls(dom, postdom);

  FOREACH_FUN_BBL(ifun, ibbl) {
    if (!ABBL_IS_MARKED(ibbl)) continue;

    if (ABBL_TYPE(ibbl) == BT_CALL) {
      cfun = ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)));
#if 1
      if (!AFUN_IS_FACTORED(cfun)) {
	return FALSE;
      }
#endif

      if (AFUN_HAS_STK_ARG(cfun)) {
	return FALSE;
      }

      if (AFUN_IS_FACTORED(cfun)) {
	if (AFUN_HAS_STK_ALIAS(ifun) &&
	    AFUN_STK_ALIASES(cfun) != REGS_NONE) {
	  return FALSE;
	}
	
	if (AFUN_HAS_STK_ARG(ifun) &&
	    AfunHasForbiddenLoads(cfun, REGSET_UNION(AFUN_STK_ALIASES(cfun), B_SP))) {
#if 0
	  printf("@@@ REJECTING STK_ARG [forbidden load: fun %d] bbl %d\n", cfun, ibbl);
#endif
	  return FALSE;
	}
      }

      iins = ABBL_INS_LAST(ibbl);
      if (AINS_CODE(iins) == I_jsr) {
	return FALSE;
      }
    }

    if (AFUN_HAS_STK_ALIAS(ifun)) {
      FOREACH_BBL_INS(ibbl, iins) {
	if (REGSET_IS_MEMBER(AINS_REG_A(iins), AFUN_STK_ALIASES(ifun))
	    || REGSET_IS_MEMBER(AINS_REG_B(iins), AFUN_STK_ALIASES(ifun))) {
	  return FALSE;
	}
	if (AINS_CODE(iins) == I_lda 
	    && AINS_REG_B(iins) == REG_SP
	    && REGSET_IS_MEMBER(AINS_REG_C(iins), AFUN_STK_ALIASES(ifun))) {
	  return FALSE;
	}
      }
    }

    if (AFUN_HAS_STK_ARG(ifun) != 0) {
      FOREACH_BBL_INS(ibbl, iins) {
	if (!(AINS_IS_LOAD(iins) || AINS_IS_STORE(iins)) ) continue;

	if (REGSET_IS_MEMBER(AINS_REG_B(iins), 
			     REGSET_UNION(AFUN_STK_ALIASES(ifun), B_SP))) {
#if 0
	  printf("@@@ REJECTING STK_ARG [forbidden load: bbl %d] ins %d\n", ibbl, iins);
#endif
	  return FALSE;
	}
      }
    }
  }

  FOREACH_BBL_INS(dom, iins) {
    if (AINS_REG_C(iins) == REG_SP) {
      return FALSE;
    }
  }
  FOREACH_BBL_INS(postdom, iins) {
    if (AINS_REG_C(iins) == REG_SP) {
      return FALSE;
    }
  }
#if 0
  if (AFUN_HAS_STK_ARG(ifun) != 0) { printf("HURRAH: passing stk-arg region <%d,%d>\n", dom, postdom); }
#endif

  return TRUE;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunPreservesRegs"

extern REGSET AfunPreservesRegs(INDEX ifun)
{
  INDEX cfun;
  INDEX ibbl;
  INDEX iins;
  REGSET regs, saveregs;

  if (AFUN_IS_VISITED(ifun)) return REGS_INT_BUT_ZERO;

  AFUN_SET_VISITED(ifun);

  regs = REGS_INT_BUT_ZERO;

  FOREACH_FUN_BBL(ifun, ibbl) {
    FOREACH_BBL_INS(ibbl, iins) {
      regs = REGSET_DIFF(regs, AinsDef(iins));
      if (AINS_CODE(iins) == I_bsr) {
	cfun = ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(AINS_BBL(iins))));
	regs = REGSET_INTERSECT(regs, AfunPreservesRegs(cfun));
      }
    }
    if (regs == REGS_NONE) return regs;
  }      

  saveregs = REGSET_INTERSECT(SAVE_REGS, AbblSavesRegs(AFUN_BBL_FIRST(ifun)));

  regs = REGSET_UNION(regs, saveregs);

  return regs;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AregionPreservesRegs"

extern REGSET AregionPreservesRegs(INDEX master)
{
  INDEX jrgn;
  REGSET regs, non_saved;
  INDEX dom, postdom;
  INDEX ifun, cfun;
  INDEX ibbl;
  INDEX iins;

  FOREACH_LIVE_FUN(ifun) {
    AFUN_UNSET_VISITED(ifun);
  }

  regs = REGS_INT_BUT_ZERO;

  for (jrgn = master; VALID(jrgn); jrgn = AREGION_LINK(jrgn)) {
    dom = AREGION_DOM(jrgn);
    postdom = AREGION_POSTDOM(jrgn);

    ifun = ABBL_FUN(dom);
    non_saved = REGSET_COMPL(AbblSavesRegs(AFUN_BBL_FIRST(ifun)));
    regs = REGSET_DIFF(regs, non_saved);

    AregionMarkBbls(dom, postdom);

    FOREACH_FUN_BBL(ifun, ibbl) {
      if (!ABBL_IS_MARKED(ibbl)) continue;
      if (ABBL_TYPE(postdom) == BT_CALL && ibbl == postdom) continue;

      FOREACH_BBL_INS(ibbl, iins) {
	regs = REGSET_DIFF(regs, AinsDef(iins));
	if (AINS_CODE(iins) == I_bsr) {
	  cfun = ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(AINS_BBL(iins))));
	  regs = REGSET_INTERSECT(regs, AfunPreservesRegs(cfun));
	}
      }
      regs = REGSET_DIFF(regs, AbblRegsLiveIn(ibbl));
      if (regs == REGS_NONE) return regs;
    }
  }

  return regs;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AregionKillOverlaps"

extern void AregionKillOverlaps(INDEX jrgn)
{
  INDEX xrgn;
  INDEX ifun;
  INDEX ibbl;
  INDEX dom, postdom, xdom, xpostdom;

  CASSERT(VALID(jrgn), ("bad argument\n"));
  CASSERT(VALID(AREGION_MASTER(jrgn)), ("invalid master: jrgn %d\n", jrgn));

  dom = AREGION_DOM(jrgn);
  postdom = AREGION_POSTDOM(jrgn);

  ifun = ABBL_FUN(dom);

  FOREACH_FUN_BBL(ifun, ibbl) {
    ABBL_UNMARK(ibbl);
  }

  FOREACH_FUN_REGION(ifun, xrgn) {
    if (!VALID(AREGION_MASTER(xrgn))) continue;
    if (xrgn == jrgn) continue;      
    
    xdom = AREGION_DOM(xrgn);
    xpostdom = AREGION_POSTDOM(xrgn);
    
    if ( (Dominates(dom, xdom) && PostDominates(postdom, xdom)) 
	 || (Dominates(dom, xpostdom) && PostDominates(postdom, xpostdom)) ) {
      if (dom == xpostdom && !ABBL_IS_MARKED(dom)) {
	/* should recompute AREGION_NINS(xrgn) */
	AREGION_POSTDOM(xrgn) = InsertBblBeforeFancy(dom);
	ABBL_MARK(dom);
      }
      else if (postdom == xdom && !ABBL_IS_MARKED(postdom)) {
	/* should recompute AREGION_NINS(xrgn) */
	AREGION_DOM(xrgn) = InsertBblAfterFancy(postdom);
	ABBL_MARK(postdom);
      }
      else {
	AREGION_MASTER(xrgn) = INVALID;   /* basically, put it out of consideration */
      }
    }
  }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorSlaveRegions"

extern void FactorSlaveRegions(INDEX master, INDEX newfun, REG retreg, NUMBER *nrgns, NUMBER *nbbls, NUMBER *nins, NUMBER savings)
{
  INDEX jrgn;
  INDEX dom, postdom;
  INDEX ibbl, jbbl, cbbl, nextbbl;
  INDEX iedg;
  INDEX iins;
  INDEX targetfun = INVALID;
  BOOL  tailcall_region;

  tailcall_region = (ABBL_TYPE(AREGION_POSTDOM(master)) == BT_CALL);

  if (tailcall_region) {
    targetfun = ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(AREGION_POSTDOM(master))));
  }

  cbbl = AFUN_BBL_FIRST(newfun);
    
  for (jrgn = master; VALID(jrgn); jrgn = AREGION_LINK(jrgn)) {
    if (!VALID(AREGION_MASTER(jrgn))) continue;
    dom = AREGION_DOM(jrgn);
    postdom = AREGION_POSTDOM(jrgn);

#if 0
    printf("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ BEFORE @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
    PrintRegion(dom, postdom);
    printf("--------------------------------------------------------------------\n");
    PrintFunSuper(stdout, ABBL_FUN(dom));
    printf("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
#endif

    AregionKillOverlaps(jrgn);

    ibbl = InsertBblBeforeFancy(dom);
    jbbl = InsertBblAfterFancy(postdom);

    iins = AinsNew();
    AinsMakeBsr(iins, retreg);
    AINS_BBL(iins) = ibbl;
    AinsAppend(iins);
    ABBL_TYPE(ibbl) = BT_CALL;  

    AedgKillFancy(ABBL_EDG_NORM(ibbl));    /* kill edge ibbl -> dom */
    AedgKillFancy(ABBL_EDG_NORM(postdom)); /* kill edge postdom -> jbbl */

    AedgLinkFancy(AedgNewFancy(ibbl, jbbl, ET_FUN_LINK));
    iedg = AedgNewFancy(ibbl, cbbl, ET_FUN_CALL);
    AedgLinkFancy(iedg);

    if (tailcall_region) {
      /* 
       * The return edge coming into jbbl needs to be changed from the exit
       * block of newfun to the exit block of targetfun.
       */
      AEDG_HEAD(iedg+1) = AFUN_BBL_LAST(targetfun);
      /*
       * right now, the out-edges from jbbl are the edges that used to be out-edges
       * from postdom, i.e., a call-edge and a link-edge.  These have to be replaced
       * by a normal edge.  
       */
      nextbbl = AEDG_TAIL(ABBL_EDG_LINK(jbbl));
      AbblKillSuccEdges(jbbl);
      /*
       * jbbl inherited the type of postdom (i.e., BT_CALL) when it was created.
       * Fix this.
       */
      ABBL_TYPE(jbbl) = BT_NORM;
      AedgLinkFancy(AedgNewFancy(jbbl, nextbbl, ET_NORM));
    }

#if 0
    printf("@@@ ADDED FUNCALL: region <%d,%d> [master <%d,%d>] now calls function %d (retreg %d)\n", dom, postdom, AREGION_DOM(master), AREGION_POSTDOM(master), newfun, retreg);
#endif
    if (jrgn != master) {
      *nrgns += 1;
      *nbbls += AREGION_NBBLS(jrgn);
      *nins += savings;
    }
#if 0
    printf("@@@@@@@@@@@@@@@@@@@@@@@@@@ AFTER: jrgn <%d,%d> @@@@@@@@@@@@@@@@@@@@@@@@@@\n", dom, postdom);
    printf("////////////////////////////// NEW FUN //////////////////////////////\n");
    PrintFunSuper(stdout, newfun);
    printf("----------------------------------------------------------------------\n");
    PrintFunSuper(stdout, ABBL_FUN(dom));
#endif
  }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorRegionNoStkAdjust"

extern void FactorRegionNoStkAdjust(INDEX master, REGSET regs, NUMBER *nrgns, NUMBER *nbbls, NUMBER *nins)
{
  INDEX jrgn;
  INDEX dom, postdom;
  INDEX ifun, newfun;
  INDEX cbbl, retbbl;
  INDEX targetbbl = INVALID;
  INDEX targetfun = INVALID;
  INDEX iins, jins;
  REG retreg;
  BOOL tailcall_region;
  int nmembers = 0, addins = 0, delins = 0, savings = 0;
  BOOL verbose = HasOption(PRINT_OPT_VERBOSE, "FactorRegion");
  BOOL xverbose = HasOption(PRINT_OPT_VERBOSE, "FuncNames");

  CASSERT(VALID(master), ("bad argument\n"));
  CASSERT(REGSET_INTERSECT(REGS_INT, regs) != REGS_NONE, ("no return reg available!\n"));

  if (AregionCallsSetjmp(master)) return;
  /*
   * if AREGION_MASTER(master) is INVALID, it means that master was killed
   * at an earlier point.  In this case we try to find another viable master
   * by following the chain of regions associated with master, in the hope
   * that something may still be salvaged.
   */
  for (jrgn = master; !VALID(AREGION_MASTER(jrgn)); jrgn = AREGION_LINK(jrgn))
    ;
  if (!VALID(jrgn)) return;
  if (jrgn != master) {
    master = jrgn;
    for ( ; VALID(jrgn); jrgn = AREGION_LINK(jrgn)) {
      if (VALID(AREGION_MASTER(jrgn))) {
	AREGION_MASTER(jrgn) = master;
      }
    }
  }

  dom = AREGION_DOM(master);
  postdom = AREGION_POSTDOM(master);
  ifun = ABBL_FUN(dom);

  for (jrgn = master; VALID(jrgn); jrgn = AREGION_LINK(jrgn)) {
    if (!VALID(AREGION_MASTER(jrgn))) continue;
    nmembers++;
  }

  addins = nmembers*1 + AREGION_NINS(master) + 1;
  delins = nmembers*AREGION_NINS(master);
  savings = delins-addins;

  if (savings <= 0) return;

  if (verbose) {
    printf("@@@ FACTORED [NOSTK:%c] REGION <%d, %d> (rec = %d) {", 
	   ABBL_TYPE(AREGION_POSTDOM(master)),
	   AREGION_DOM(master), AREGION_POSTDOM(master),
	   AFUN_IS_RECURSIVE(AFUN_IS_RECURSIVE(ABBL_FUN(AREGION_DOM(master)))));
    for (jrgn = AREGION_LINK(master); VALID(jrgn); jrgn = AREGION_LINK(jrgn)) {
      printf(" <%d, %d> (rec = %d) ", AREGION_DOM(jrgn), AREGION_POSTDOM(jrgn),
	     AFUN_IS_RECURSIVE(ABBL_FUN(AREGION_DOM(jrgn))));
    }
    printf(" } (%d ins) [add %d, del %d, savings %d]\n", AREGION_NINS(master),
	   addins, delins, savings);
  }

  newfun = CopyRegion2Fun(master);
  cbbl = AFUN_BBL_FIRST(newfun);
  /*
   * find a suitable register for the return address, and fix the return block
   * of the newly created function accordingly.
   */
  retreg = FindRetReg(regs);
  retbbl = ABBL_MAGIC(postdom);
  
  tailcall_region = (ABBL_TYPE(postdom) == BT_CALL);

  switch (ABBL_TYPE(postdom)) {  
  case BT_NORM:
    iins = AinsNew();
    AinsMakeRet(iins, retreg);
    AINS_BBL(iins) = retbbl;
    AinsAppend(iins);
    break;
    
  case BT_CALL:
    iins = ABBL_INS_FIRST(retbbl);
    CASSERT(!VALID(AINS_NEXT(iins)),
	    ("Control flow separation assumption violated: bbl %d\n", retbbl));
    CASSERT(AINS_CODE(iins) == I_bsr, 
	    ("non-bsr call instruction in post-dominator: %d\n", retbbl));
    
    targetbbl = AEDG_TAIL(ABBL_EDG_CALL(postdom));
    targetfun = ABBL_FUN(targetbbl);

    if (retreg != AINS_REG_C(iins)) {
      jins = AinsNew();
      AinsMakeLda(jins, retreg, AINS_REG_C(iins), 0);
      AINS_BBL(jins) = retbbl;
      AinsAppend(jins);
    }
    
    AinsKill(iins);
    
    ABBL_TYPE(retbbl) = BT_NORM;
    AbblKillSuccEdges(retbbl);
    AedgLinkFancy(AedgNewFancy(retbbl, targetbbl, ET_NORM));
    break;
    
  default:
    ERROR(("unexpected post-dom type %c\n", ABBL_TYPE(postdom)));
    break;
  }
  /*
   * Now replace each occurrence of the region by a call to the new function.
   */
  if (xverbose) {
    for (jrgn = master; VALID(jrgn); jrgn = AREGION_LINK(jrgn)) {
      if (!VALID(AREGION_MASTER(jrgn))) continue;
      printf("@@@FUNC   %s   ins   %.2f\n", AFUN_NAME(ABBL_FUN(AREGION_DOM(jrgn))), 
	     (double)savings/(double)nmembers);
    }
  }

  FactorSlaveRegions(master, newfun, retreg, nrgns, nbbls, nins, savings);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorRegionStkAdjust"

extern void FactorRegionStkAdjust(INDEX master, NUMBER *nrgns, NUMBER *nbbls, NUMBER *nins)
{
  INDEX jrgn;
  INDEX dom, postdom;
  INDEX ifun, newfun, targetfun;
  INDEX cbbl, retbbl, targetbbl;
  INDEX iins;
  REGSET liveregsin, liveregsout;
  REG retreg0, retreg1;
  BOOL tailcall_region;
  int nmembers = 0, addins = 0, delins = 0, savings = 0;
  BOOL verbose = HasOption(PRINT_OPT_VERBOSE, "FactorRegion");
  BOOL xverbose = HasOption(PRINT_OPT_VERBOSE, "FuncNames");

  CASSERT(VALID(master), ("bad argument\n"));

  if (AregionCallsSetjmp(master)) return;
  /*
   * if AREGION_MASTER(master) is INVALID, it means that master was killed
   * at an earlier point.  In this case we try to find another viable master
   * by following the chain of regions associated with master, in the hope
   * that something may still be salvaged.  
   */
  for (jrgn = master; !VALID(AREGION_MASTER(jrgn)); jrgn = AREGION_LINK(jrgn))
    ;
  if (!VALID(jrgn)) return;
  if (jrgn != master) {
    master = jrgn;
    for ( ; VALID(jrgn); jrgn = AREGION_LINK(jrgn)) {
      if (VALID(AREGION_MASTER(jrgn))) {
	AREGION_MASTER(jrgn) = master;
      }
    }
  }

  dom = AREGION_DOM(master);
  postdom = AREGION_POSTDOM(master);

#if 1
  if (ABBL_TYPE(postdom) == BT_CALL) return;
#endif

  liveregsin = SAVE_REGS;
  liveregsout = SAVE_REGS;
  for (jrgn = master; VALID(jrgn); jrgn = AREGION_LINK(jrgn)) {
    if (!VALID(AREGION_MASTER(jrgn))) continue;

    nmembers++;
    liveregsin = REGSET_UNION(liveregsin, AbblRegsLiveIn(AREGION_DOM(jrgn)));
    liveregsout = REGSET_UNION(liveregsout, ABBL_OUT(AREGION_POSTDOM(jrgn)));
  }

  if (REGSET_IS_SUBSET(REGS_INT_BUT_ZERO, liveregsin)) {
    return;
  }
  if (ABBL_TYPE(postdom) != BT_CALL 
      && REGSET_IS_SUBSET(REGS_INT_BUT_ZERO, liveregsout)) {
    return;
  }

  tailcall_region = (ABBL_TYPE(postdom) == BT_CALL);

  if (tailcall_region) {
#if 0
    printf("@@@ TAILCALL: region <%d,%d>\n", AREGION_DOM(jrgn), AREGION_POSTDOM(jrgn));
#endif
    FOREACH_BBL_INS_R(postdom, iins) {
      if (AINS_CODE(iins) == I_bsr) break;
    }
    CASSERT(VALID(iins), ("call block does not end in bsr: %d\n", postdom));
    retreg1 = AINS_REG_C(iins);
  }
  else {
    retreg1 = FindRetReg(REGSET_COMPL(liveregsout));
  }

  addins = nmembers*1 + AREGION_NINS(master) + 5;
  delins = nmembers*AREGION_NINS(master);
  savings = delins-addins;

  if (retreg1 == REG_RA) {
    addins -= 1;  /* restore reg factoring will get rid of one instr */
  }

  if (savings <= 0) return;

  ifun = ABBL_FUN(dom);
  CASSERT((AFUN_HAS_STK_ALIAS(ifun) == 0), ("function %d has stack alias [region <%d,%d>]\n", ifun, dom, postdom));
  CASSERT((AFUN_HAS_STK_INVALID(ifun) == 0), ("function %d has stack invalid [region <%d,%d>]\n", ifun, dom, postdom));
  CASSERT((AFUN_HAS_STK_ARG(ifun) == 0), ("function %d has stack arg [region <%d,%d>]\n", ifun, dom, postdom));

  newfun = CopyRegion2Fun(master);

  FactoredRegionAddPreHeader(newfun);

  cbbl = AFUN_BBL_FIRST(newfun);

  AfunShiftStackRefs(newfun, 8);

  retreg0 = FindRetReg(REGSET_COMPL(liveregsin));
    
  iins = AinsNew();
  AinsMakeStq(iins, retreg0, REG_SP, 0);
  AINS_BBL(iins) = cbbl;
  AinsPrepend(iins);

  retbbl = ABBL_MAGIC(postdom);
  
  iins = AinsNew();
  AinsMakeLda(iins, REG_SP, REG_SP, -8);
  AINS_BBL(iins) = cbbl;
  AinsPrepend(iins);

  targetfun = INVALID;  /* to keep the compiler happy */
  switch (ABBL_TYPE(postdom)) {  
  case BT_NORM:
    iins = AinsNew();
    AinsMakeLdq(iins, REG_SP, retreg1, 0);
    AINS_BBL(iins) = retbbl;
    AinsAppend(iins);

    iins = AinsNew();
    AinsMakeLda(iins, REG_SP, REG_SP, 8);
    AINS_BBL(iins) = retbbl;
    AinsAppend(iins);

    iins = AinsNew();
    AinsMakeRet(iins, retreg1);
    AINS_BBL(iins) = retbbl;
    AinsAppend(iins);
    break;
    
  case BT_CALL:
    iins = ABBL_INS_FIRST(retbbl);
    CASSERT(!VALID(AINS_NEXT(iins)),
	    ("Control flow separation assumption violated: bbl %d\n", retbbl));
    CASSERT(AINS_CODE(iins) == I_bsr, 
	    ("non-bsr call instruction in post-dominator: %d\n", retbbl));
    
    targetbbl = AEDG_TAIL(ABBL_EDG_CALL(postdom));
    targetfun = ABBL_FUN(targetbbl);

    AinsKill(iins);

    iins = AinsNew();
    AinsMakeLdq(iins, REG_SP, retreg1, 0);
    AINS_BBL(iins) = retbbl;
    AinsAppend(iins);

    iins = AinsNew();
    AinsMakeLda(iins, REG_SP, REG_SP, 8);
    AINS_BBL(iins) = retbbl;
    AinsAppend(iins);
    
    ABBL_TYPE(retbbl) = BT_NORM;
    AbblKillSuccEdges(retbbl);
    AedgLinkFancy(AedgNewFancy(retbbl, targetbbl, ET_NORM));
    AedgLinkFancy(AedgNewFancy(retbbl, AFUN_BBL_LAST(newfun), ET_EXIT_LINK));
    break;
    
  default:
    ERROR(("unexpected post-dom type %c\n", ABBL_TYPE(postdom)));
    break;
  }

  if (verbose) {
    printf("@@@ FACTORED REGION[%c] <%d, %d> (rec = %d) {", 
	   ABBL_TYPE(AREGION_POSTDOM(master)),
	   AREGION_DOM(master), AREGION_POSTDOM(master),
	   AFUN_IS_RECURSIVE(ABBL_FUN(AREGION_DOM(master))));
    for (jrgn = AREGION_LINK(master); VALID(jrgn); jrgn = AREGION_LINK(jrgn)) {
      printf(" <%d, %d> (rec = %d) ", AREGION_DOM(jrgn), AREGION_POSTDOM(jrgn),
	     AFUN_IS_RECURSIVE(ABBL_FUN(AREGION_DOM(jrgn))));
    }
    printf(" } (%d ins) [add %d, del %d, savings %d]\n", AREGION_NINS(master),
	   addins, delins, savings);
  }

  if (xverbose) {
    for (jrgn = master; VALID(jrgn); jrgn = AREGION_LINK(jrgn)) {
      if (!VALID(AREGION_MASTER(jrgn))) continue;
      printf("@@@FUNC   %s   ins   %.2f\n", AFUN_NAME(ABBL_FUN(AREGION_DOM(jrgn))), 
	     (double)savings/(double)nmembers);
    }
  }

  FactorSlaveRegions(master, newfun, retreg0, nrgns, nbbls, nins, savings);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorRegionTCO"

extern void FactorRegionTCO(INDEX master, NUMBER *nrgns, NUMBER *nbbls, NUMBER *nins)
{
  INDEX jrgn;
  INDEX ifun;
  INDEX ebbl, dom, postdom, xbbl;
  INDEX ibbl, iedg;
  BOOL verbose = FALSE;
  BOOL xverbose = HasOption(PRINT_OPT_VERBOSE, "FuncNames");

  CASSERT(VALID(master), ("bad argument\n"));

  AregionMarkBbls(AREGION_DOM(master), AREGION_POSTDOM(master));

  ebbl = AREGION_DOM(master);
  ifun = ABBL_FUN(ebbl);

  if ( xverbose ) {
    for (jrgn = master; VALID(jrgn); jrgn = AREGION_LINK(jrgn)) {
      printf("@@@FUNC   %s   ins   %d\n", AFUN_NAME(ABBL_FUN(AREGION_DOM(jrgn))), 
	     AREGION_NINS(jrgn));
    }
  }

  for (jrgn = AREGION_LINK(master); VALID(jrgn); jrgn = AREGION_LINK(jrgn)) {
    if (!VALID(AREGION_MASTER(jrgn))) continue;

    dom = AREGION_DOM(jrgn);
    postdom = AREGION_POSTDOM(jrgn);
    CASSERT((ABBL_TYPE(postdom) == BT_RET), ("postdom not of type BT_RET\n"));

    ibbl = InsertBblBeforeFancy(dom);
    iedg = ABBL_EDG_NORM(ibbl);
    AedgKillFancy(iedg);
    AedgLinkFancy(AedgNewFancy(ibbl, ebbl, ET_NORM));
    
    FOREACH_FUN_BBL(ifun, xbbl) {
      if (!ABBL_IS_MARKED(xbbl)) continue;

      ABBL_COUNT(xbbl) += ABBL_COUNT(dom);
    }
    
    AregionKillOverlaps(jrgn);
    AREGION_MASTER(jrgn) = INVALID;  /* to avoid processing it again */
    
    *nrgns += 1;
    *nbbls += AREGION_NBBLS(jrgn);
    *nins += AREGION_NINS(jrgn);
    
    if (verbose) {
      printf("@@@ REGION [R: <%d,%d> -> <%d,%d> (%d bbls, %d instrs)]: added edge %d -> %d\n",
	     dom, postdom, ebbl, AREGION_POSTDOM(master), 
	     AREGION_NBBLS(jrgn), AREGION_NINS(jrgn), 
	     ibbl, ebbl);
    }
  }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorRegions"

extern void FactorRegions()
{
  INDEX irgn;
  INDEX master;
  INDEX dom, postdom;
  REGSET regs;
  NUMBER nrgns = 0, nbbls = 0, nins = 0;
  int i;
  clock_t time = clock();
  BOOL verbose = HasOption(PRINT_OPT_VERBOSE, "FactorRegion");

  AfunMarkRecursiveFuns();
  ComputeRegions();
  GroupIdenticalRegions();

  for (i = maxregion; i >= minregion; i--)
    {
      FOREACH_REGION(irgn)
	{
	  if (AREGION_NBBLS(irgn) != i) continue;
	  if (!VALID(AREGION_MASTER(irgn))) continue;
	  if (AREGION_MASTER(irgn) != irgn) continue;
	  
	  master = AREGION_MASTER(irgn);
	  
	  dom = AREGION_DOM(master);
	  postdom = AREGION_POSTDOM(master);
	  if ( !(Dominates(dom,postdom) && PostDominates(postdom,dom)) ) continue;
	  
	  if (verbose)
	    {
	      printf("@@@ FactorRegion: dom %d [ fun %d ], postdom %d [ fun %d ] type %c\n",
		     dom, ABBL_FUN(dom), postdom, ABBL_FUN(postdom), ABBL_TYPE(postdom));
	    }
	  
	  switch (ABBL_TYPE(postdom)) 
	    {
	    case BT_RET:
	      FactorRegionTCO(master, &nrgns, &nbbls, &nins);
	      break;
	    case BT_NORM:
#if 0
	    case BT_CALL:
#endif
	      regs = AregionPreservesRegs(master);
	      if (REGSET_INTERSECT(REGS_INT,regs) != REGS_NONE) 
		{
		  FactorRegionNoStkAdjust(master, regs, &nrgns, &nbbls, &nins);
		}
	      else if (ARegionIsWellBehaved(master)) 
		{
		  FactorRegionStkAdjust(master, &nrgns, &nbbls, &nins);
		}
	      break;
	      
	    default:
	      continue;
	    }
	  AREGION_MASTER(master) = AREGION_LINK(master) = INVALID;
	}
    }
  
  time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
  InfoStatus(stdout,"\t[FactorRegions]: savings %d (%d regions, %d bbls) { %d ms }\n",
	     nins, nrgns, nbbls, time);
}

static void FactoredRegionAddPreHeader(INDEX ifun)
{
  INDEX ibbl;
  INDEX jbbl;

  if (AbblNpreds(AFUN_BBL_FIRST(ifun))==0) return;

  ibbl= AFUN_BBL_FIRST(ifun);
  
  jbbl = AbblNew();

  ABBL_COUNT(jbbl)=ABBL_COUNT(ibbl);
  ABBL_TYPE(jbbl)=BT_NORM;
  ABBL_FUN(jbbl)=ifun;

  AbblAppendFirst(jbbl);
  
  AedgLinkFancy(AedgNewFancy(jbbl,ibbl,ET_NORM));  
}

